
#define _GNU_SOURCE

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <errno.h>

#include <microhttpd.h>

#include "packet.h"
#include "log.h"

#include "udp_server.h"

#include "SamsungDmsDesc.h"

      char *source      = NULL;
const char *destination = NULL;

static int upnp_udp_receive(const struct sockaddr src_addr, const uint8_t * const buffer, const int length) {
	// see if the UPnP response is for us
	if(!pmatch(buffer, 346)) return(1);

	// replace IP adres with our interface address
	uint8_t out[512];
	const int len = preplace(349, out, "192.168.12.11:52235", source);

	// transmit back UPnP reponse
	udp_send(&src_addr, 1819, out, len);
	return(1);
}

char notfound[] = "<html><head><title>404 Not Found!</title></head><body><h1>404 Not Found!</h1></body></html>";
char ise[] =      "<html><head><title>500 Internal server error!</title></head><body><h1>500 Internal server error!</h1></body></html>";

typedef struct {
	// accumulates POST data
	char *post_data;
	size_t post_size;
} my_conn_t;

int answer_to_connection(void *cls, struct MHD_Connection *connection, const char *url, const char *method, const char *version, const char *upload_data, size_t *upload_data_size, void **con_cls) {
	char buffer[2048];
	char *page = notfound;
	int   len  = strlen(page);

	if(! *con_cls) {
		// assign empty memory for handling this context
		*con_cls = calloc(1, sizeof(my_conn_t));
		if(! *con_cls) syslog(LOG_CRIT, "answer_to_connection: calloc failed: %s", strerror(errno));

		// the first time only the headers are valid, do not respond
		return(MHD_YES);
	}

	// capture back my context
	my_conn_t *conn = (my_conn_t*)*con_cls;

	// if upload_data_size is not null keep on downloading
	if(*upload_data_size) {
		const int post_idx = conn->post_size;

		// recalculate our new buffer size 
		conn->post_size += *upload_data_size;
		
		// resize our buffer
		conn->post_data  = realloc(conn->post_data, conn->post_size);
		if(conn->post_data == NULL) syslog(LOG_CRIT, "answer_to_connection: realloc failed: %s", strerror(errno));

		// append post data to our buffre
		memcpy(conn->post_data+post_idx, upload_data, *upload_data_size);

		// let server know we have handled the data
		*upload_data_size = 0 ;
		return(MHD_YES);
	}

	// check for simple GET documents
	if(strcmp(method, "GET") == 0) {
	  if(strcmp(url, "/DMS/SamsungDmsDesc.xml") == 0) {
			page = SamsungDmsDesc;
			len  = strlen(page);
		}
	}
	
	// check for POST documents
	else if(strcmp(method, "POST") == 0 && strcmp(url, "/upnp/control/ContentDirectory1") == 0) {
		if(strstr(conn->post_data, "<s:Body><u:X_BACKUP_START ")) {
			syslog(LOG_INFO, "Backup starting...");

			page = (char*)hacket(372);
			len  = hsize(372);
		} 
		else if(strstr(conn->post_data, "<s:Body><u:X_BACKUP_DONE ")) {
			syslog(LOG_INFO, "Backup finished!");

			page = (char*)hacket(7826);
			len  = hsize(7826);
		} 
		else if(strstr(conn->post_data, "<s:Body><u:CreateObject ")) {
			char *end;

#     define field(match) (strstr(conn->post_data, match) + sizeof(match) - 1)
#     define terminate(field, term) { end = strstr(field, term); *end = '\0'; }

			const char * const title = field("dc:title&gt;");
			const char * const date = field("dc:date&gt;");
			const char * const size = field("&quot; size=&quot;");

			terminate(title, "&lt;");
			terminate(date,  "&lt;");
			terminate(size,  "&quot;");

			syslog(LOG_INFO, "Request to store foto %s/%s of %s bytes...", date, title, size);

			len = sprintf(buffer, "<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\"><s:Body><u:CreateObjectResponse xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\"><ObjectID>%s/%s</ObjectID><Result>&lt;DIDL-Lite xmlns=&quot;urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/&quot; xmlns:dc=&quot;http://purl.org/dc/elements/1.1/&quot; xmlns:upnp=&apos;urn:schemas-upnp-org:metadata-1-0/upnp/&apos; xmlns:dlna=&quot;urn:schemas-dlna-org:metadata-1-0/&quot; xmlns:sec=&quot;http://www.sec.co.kr/&quot;&gt;&lt;item id=&quot;%s/%s&quot; parentID=&quot;%s&quot; restricted=&quot;0&quot; dlna:dlnaManaged=&quot;00000004&quot;&gt;&lt;dc:title&gt;&lt;/dc:title&gt;\n&lt;res protocolInfo=&quot;http-get:*:image/jpeg:DLNA.ORG_PN=JPEG_LRG;DLNA.ORG_CI=0;DLNA.ORG_FLAGS=00D00000000000000000000000000000&quot; importUri=&quot;http://%s/upload/%s/%s&quot; dlna:resumeUpload=&quot;1&quot; dlna:uploadedSize=&quot;0&quot; size=&quot;%s&quot;&gt;&lt;/res&gt;\n&lt;upnp:class&gt;object.item.imageItem&lt;/upnp:class&gt;\n&lt;/item&gt;\n&lt;/DIDL-Lite&gt;</Result></u:CreateObjectResponse></s:Body></s:Envelope>\r\n", date, title, date, title, date, source, date, title, size);
			page = buffer; 

			// build a path to the folder to store the foto in
			char folder[strlen(destination)+1+strlen(date)+1];
			const int folder_len = sprintf(folder, "%s/%s", destination, date);

			// create the date folder to store the foto in
			// ignoring errors, we don't care if the folder already exists
			mkdir(folder, 0777);
		} 
	}

	// check for foto upload POST data, abusing for to break on error
	else if(strcmp(method, "POST") == 0 && strstr(url, "/upload/")) for(;;) {
		// capture filename of foto to store
		const char * const filename = url+8;

		// build a path to the file to store the foto in
		char path[strlen(destination)+1+strlen(filename)+1];
		sprintf(path, "%s/%s", destination, filename);

		syslog(LOG_INFO, "Writing foto to %s...", path);

		// open the foto for writing
		FILE * const fd = fopen(path, "w");
		if(fd == NULL) {
			syslog(LOG_CRIT, "answer_to_connection: fopen failed: %s", strerror(errno));

			// return internal server error page
			page = ise;
			len  = strlen(page);
			break;
		}

		// try to store the foto, if this fails we return an internal system error
		if(fwrite(conn->post_data, 1, conn->post_size, fd) < conn->post_size) {
			syslog(LOG_CRIT, "answer_to_connection: fwrite failed: %s", strerror(errno));
			fclose(fd);

			// return internal server error page
			page = ise;
			len  = strlen(page);
			break;
		}

		// write succesful, close the file
		fclose(fd);

		// return empty document on file download
		page = "";
		len  = 0;
		break;
	}

	// cleanup post memory
	if(conn->post_data != NULL) free(conn->post_data);

	// cleanup context memory
	free(*con_cls);

	// clear context pointer
	*con_cls = NULL;

	struct MHD_Response *response = MHD_create_response_from_buffer(strlen(page), page, MHD_RESPMEM_MUST_COPY);
	MHD_add_response_header(response, "CONTENT-TYPE", "text/xml; charset=\"utf-8\"");
	MHD_add_response_header(response, "SERVER", "MS-Windows/XP UPnP/1.0 PROTOTYPE/1.0");
	const int ret = MHD_queue_response(connection, page==notfound?MHD_HTTP_NOT_FOUND:(page==ise?MHD_HTTP_INTERNAL_SERVER_ERROR:MHD_HTTP_OK), response);
	MHD_destroy_response(response);

	return ret;
}

int main(int argc, char *argv[]) {
	const char * const progname = *argv;

	// use syslog for logging
	openlog(progname, LOG_NDELAY|LOG_PERROR|LOG_PID, LOG_FTP);

	// by default don't daemonize
	int daemonize = 0;

	// by default no failure
	int failed = 0;

	// default PID file to use
	const char * const _pidfile = "/var/run/dv300f.pid";
	const char * pidfile = _pidfile;

	// no default interface
	const char * interface_address = NULL;
	const char * interface = NULL;

	// default http port to use
	const int _http_port = 52235;
	int http_port = _http_port;

	// parse command line
	int opt;
	while((opt = getopt(argc, argv, "i:f:I:P:dp:")) != -1) {
		switch(opt) {
			case('i'):
				interface_address = optarg;
				break;
			case('I'):
				interface = optarg;
				break;
			case('P'):
				http_port = atoi(optarg);
				break;
			case('f'):
				destination = optarg;
				break;
			case('d'):
				daemonize = 1;
				break;
			case('p'):
				pidfile = optarg;
				break;
			default:
				// fail on invalid input
				failed = 1;
				break;
		}
	}

	// interface and destinatin must be set at all times
	if(interface_address == NULL || destination == NULL) failed = 1;

	// check for a valid command line
	if(failed) {
		syslog(LOG_ERR, "Usage: %s <-i interface_address> [-I interface] [-P http port] <-f destination_folder> [-d] [-p pidfile]", progname);
		syslog(LOG_ERR, "  -i specifies the interface address to feedback in UPnP messages");
		syslog(LOG_ERR, "  -I specifies the interface to bind on");
		syslog(LOG_ERR, "  -P specifies the http port to feedback in UPnP messages, defaults to %u", _http_port);
		syslog(LOG_ERR, "  -f is a path to the destination folder to store the photos in");
		syslog(LOG_ERR, "  -d tells this program to daemonize itself");
		syslog(LOG_ERR, "  -p specifiies the path to a PID file to use when daemonizing, defaults to %s", pidfile);
		return(1);
	}

	// need to daemonize?
	if(daemonize) {
		// stops the parent waiting for the child process to end
		signal(SIGCHLD, SIG_IGN);

		// try to fork
		pid_t pid = fork();

		// did the fork fail?
		if(pid < 0) {
			syslog(LOG_ERR, "Fork of a child failed: %s", strerror(errno));
			return(2);
		}
		
		// are we the parent?
		if(pid) {
			// open a pidfile to write our pid to
			FILE * const pidfd = fopen(pidfile, "w");

			// faile to open pidfile?
			if(pidfd == NULL) {
				syslog(LOG_ERR, "Open of the pidfile %s failed: %s", pidfile, strerror(errno));

				// kill the child while we still can
				kill(pid, 15);
				return(3);
			}

			// write pid to the pidfile
			fprintf(pidfd, "%d\n", pid);
			fclose(pidfd);

			syslog(LOG_INFO, "Succesfully daemonized, child PID is %d", pid);

			// close with success (daemon child will stay behind)
			return(0);
		}

		// let's just sit here for a second, our parent might still kill us
		sleep(1);

		// re-open the log without printing to STDERR as we have daemonized
		closelog();
		openlog(progname, LOG_NDELAY|LOG_PID, LOG_FTP);
	}

	// select helper for UDP server UPnP hack
	select_init();

	// setup multicast UDP server on 239.255.255.250:1900
	udp_server_t upnp_udp;
	if(!udp_server_open(&upnp_udp, 1900, upnp_udp_receive, interface)) return(4);
	if(!udp_server_set_mg(&upnp_udp, "239.255.255.250", interface_address)) return(5);

	// create source
	asprintf(&source, "%s:%u", interface_address, http_port);
	if(source == NULL) {
		syslog(LOG_ERR, "asprintf fialed, exiting...");
		return(7);
	}

	// setup HTTP server on port 52235
	struct MHD_Daemon *upnp_http;
	upnp_http = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY, http_port, NULL, NULL, answer_to_connection, NULL, MHD_OPTION_END);
	if(upnp_http == NULL) {
		syslog(LOG_ERR, "MHD_start_daemon failed, exiting...");
		return(6);
	}

	// let user know we are ready
	syslog(LOG_INFO, "PC auto backup ready!");

	// we never stop
	for(;;) {
		// drive UDP server (HTTP server is seperate thread)
		select_tick(-1);
	}
}
